module doubleLinkedList
    ! Dieses Modul definiert eine Integer-Liste, welche als doppelt verkettete
    ! Liste implementiert ist. Dazu werden die unten stehenden Funktionalitäten 
    ! bereitgestellt.
    implicit none
    private 
    public :: list, push, pop, inject, eject, put, deleteAll, getItem, traverse 
    
    type listElem! wird nicht exportiert, nur innerhalb des Moduls verfügbar!
        integer :: content
        type(listElem), pointer :: prev => null(), next => null()
    end type listElem
    type list 
        type(listElem), pointer :: head => null(), tail => null() 
        integer :: length = 0
    end type list
    
 contains
    
    subroutine putInteger(i)! Ausgabe integer
        integer, intent(in) :: i
        write(*,"(I6)",advance="no") i
    end subroutine

    logical function isEmpty(L)! testen, ob Liste leer
        type(list), intent(in) :: L
        isEmpty = .not. associated(L%head)
    end function

    integer function length(L)! Länge der Liste ermitteln
        type(list), intent(in) :: L 
        length = L%length
    end function

    subroutine push(L, newData)! neuen Listeneintrag am Anfang anhängen
        type(List), intent(inout) :: L 
        integer, intent(in) :: newData
        type(listElem), pointer :: tmp
        allocate(tmp)
        tmp%content = newData
        
        if ( isEmpty(L) ) then
            L%tail => tmp
        else 
            tmp%next => L%head
            L%head%prev => tmp
        end if
        
        L%head => tmp
        L%length = L%length + 1
    end subroutine

    subroutine inject(L, newData)! neuen Listeneintrag am Ende anhängen
        type(List), intent(inout) :: L 
        integer, intent(in) :: newData
        type(listElem), pointer :: tmp
        allocate(tmp)
        tmp%content = newData
        
        if ( isEmpty(L) ) then
            L%head => tmp
        else 
            tmp%prev => L%tail
            L%tail%next => tmp
        end if
        
        L%tail => tmp
        L%length = L%length + 1
    end subroutine

    subroutine pop(L, saveData)! ersten Listeneintrag löschen (ggf. dabei speichern)
        type(list), intent(inout) :: L
        integer, intent(out), optional :: saveData
        type(listElem), pointer :: tmp
        if ( isEmpty(L) ) then
            write(*,*) "Warning: Cannot push from empty List"
        else
            if ( present(saveData) ) then
                saveData = L%head%content
            end if
            
            if ( length(L) == 1 ) then
                deallocate(L%head)
                L%tail => null()
            else 
                tmp => L%head 
                L%head => L%head%next 
                L%head%prev => null()
                deallocate(tmp)
            end if
            L%length = L%length - 1
        end if
    end subroutine

    subroutine eject(L, saveData)! letzten Listeneintrag löschen (ggf. dabei speichern)
        type(list), intent(inout) :: L
        integer, intent(out), optional :: saveData
        type(listElem), pointer :: tmp
        if ( isEmpty(L) ) then
            write(*,*) "Warning: Cannot eject from empty List"
        else
            if ( present(saveData) ) then
                saveData = L%tail%content
            end if
            
            if ( length(L) == 1 ) then
                deallocate(L%head)
                L%tail => null()
            else 
                tmp => L%tail 
                L%tail => L%tail%prev 
                L%tail%next => null()
                deallocate(tmp)
            end if
            L%length = L%length - 1
        end if
    end subroutine

    subroutine put(L)! Ausgabe der Liste
        type(list), intent(in) :: L
        type(listElem), pointer :: tmp
        integer :: i
        tmp => L%head 
        write(*,"(A)", advance="no") "["
        do i = 1, L%length - 1
            call putInteger(tmp%content)
            write(*,"(A)", advance="no") ", "
            tmp => tmp%next 
        end do 
        if (associated(tmp)) then 
            call putInteger(tmp%content)
        end if
        write(*,"(A)") "]"
    end subroutine

    subroutine deleteAll(L)! Löschen der Listenelemente
        type(list), intent(inout) :: L
        do while(.not. isEmpty(L))
            call pop(L)
        end do 
    end subroutine

    subroutine getItem(L, i, saveData)! i-ten Listeneintrag auslesen
        type(list), intent(in) :: L
        type(listElem), pointer :: tmp
        integer, intent(out) :: saveData
        integer, intent(in) :: i
        integer :: k 
        if ( i > 0 .and. i<=L%length ) then
            tmp => L%head
            do k=1,i-1
                tmp => tmp%next
            end do 
            saveData = tmp%content
        else 
            write(*,*) "Index out of bounds in getItem."
        end if
    end subroutine

    subroutine traverse(L, task)! Über die Liste iterieren und eine "task" Subroutine
                                ! für jeden Listeneintrag ausführen (ohne die Liste
                                ! oder ihre Einträge zu verändern)
        type(list), intent(in) :: L
        type(listElem), pointer :: tmp
        integer :: i
        interface 
            subroutine task(someData)
                integer, intent(in) :: someData
            end subroutine
        end interface 
        tmp => L%head
        do i = 1, L%length
            call task(tmp%content)
            tmp => tmp%next
        end do 
    end subroutine
end module doubleLinkedList

module highLevelListFunctions
    ! Dieses Modul implementiert einige "High-Level"-Funktionalitäten für Listen.
    ! Man beachte, dass keine Pointer verwendet werden müssen!
    ! Alle drei Funktionalitäten werden zunächst als Subroutinen implementiert
    ! (denn dies ermöglicht die Nutzung von Unter-Subroutinen, sodass letztendlich
    !  "traverse" verwendet werden kann). Dann werden entsprechende Funktionen
    ! definiert und nur diese implementiert (da nur diese gewünscht waren).
    use doubleLinkedList
    implicit none
    private 
    public :: map, foldLeft, filter
contains
    ! 1. MAP:
    subroutine mapSubroutine(inputList, someFunction, outputList)
        type(list), intent(in) :: inputList
        type(list), intent(out) :: outputList
        interface
            integer function someFunction(someData)
                integer, intent(in) :: someData
            end function someFunction
        end interface
        call traverse(inputList,taskForMap)! Führe für alle Listeneinträge taskForMap aus
    contains
        subroutine taskForMap(someData) 
            ! TaskForMap: Funktionswert von someFunction am Ende der outputList einfügen
            integer, intent(in) :: someData
            call inject(outputList, someFunction(someData))
        end subroutine taskForMap
    end subroutine mapSubroutine

    ! Mittels der Subroutine mapSubroutine die Funktion map definieren
    function map(inputList,someFunction) result(outputList) 
        type(list), intent(in) :: inputList
        type(list) :: outputList
        interface
            integer function someFunction(someData)
                integer, intent(in) :: someData
            end function someFunction
        end interface
        call mapSubroutine(inputList,someFunction,outputList)
    end function map

    ! 2. FOLDLEFT
    subroutine foldLeftSubroutine(someList, someFunction, someStartValue, outputValue) 
        type(list), intent(in) :: someList
        integer, intent(in) :: someStartValue
        integer, intent(out) :: outputValue
        interface
            integer function someFunction(someData, otherData)
                integer, intent(in) :: someData, otherData
            end function someFunction
        end interface
        outputValue = someStartValue! outputValue initialisieren
        call traverse(someList, taskFoldLeft)! Für alle Listeneinträge taskFoldLeft
                                             ! ausführen.
    contains                                 
        subroutine taskFoldLeft (someData)
            ! taskFoldLeft: outputValue mittels someFunction und someData (entspricht
            ! jeweiligem Listeneintrag) aktualisieren
            integer, intent(in) :: someData
            outputValue = someFunction(outputValue,someData)
        end subroutine taskFoldLeft
    end subroutine foldLeftSubroutine

    ! Mittels foldLeftSubroutine die Funktion foldLeft definieren
    integer function foldLeft(someList,someFunction,someStartValue) result(outputValue)
        type(list), intent(in) :: someList
        integer, intent(in) :: someStartValue
        interface
            integer function someFunction(someData, otherData)
                integer, intent(in) :: someData, otherData
            end function someFunction
        end interface
        call foldLeftSubroutine(someList,someFunction,someStartValue,outputValue)
    end function foldLeft

    ! 3. FILTER
    subroutine filterSubroutine(inputList,someCondition,outputList)
        type(list), intent(in) :: inputList
        type(list), intent(out) :: outputList
        interface
            logical function someCondition(someData)
                integer, intent(in) :: someData
            end function someCondition
        end interface
        call traverse(inputList,taskFilter)! Für alle Listeneinträge taskFilter ausführen
    contains 
        subroutine taskFilter(someData)
            ! taskFilter: Wenn someCondition erfüllt, dann in outputList hinzufügen
            integer, intent(in) :: someData
            if ( someCondition(someData) ) then
                call inject(outputList, someData)
            end if
        end subroutine taskFilter
    end subroutine filterSubroutine

    ! Mittels filterSubroutine die Funktion filter definieren
    function filter(inputList, someCondition) result(outputList)
        type(list), intent(in) :: inputList
        type(list) :: outputList
        interface
            logical function someCondition(someData)
                integer, intent(in) :: someData
            end function someCondition
        end interface
        call filterSubroutine(inputList, someCondition, outputList)
    end function filter
end module highLevelListFunctions

program main
    use doubleLinkedList
    use highLevelListFunctions
    implicit none
    type(list) :: L
    integer :: i
    
    ! Liste aufbauen
    do i=1,5
        call push(L, i)
        call inject(L, 10*i)
    end do
    call put(L)! Ausgabe
    
    call put(map(L,square))! Ausgabe Quadrate der ursprünglichen Listeneinträge

    write(*,*) foldleft(L,add,0)! Ausgabe Summe der Listeneinträge

    call put(filter(L,even))! Ausgabe der geraden Listeneinträge
    
    call deleteAll(L)! Liste löschen
    ! ACHTUNG: Die durch map und filter neu entstandenen Listen wurden nicht gelöscht!
    ! Lösung: Ergebnislisten von map und filter einer Variable zuweisen.
    ! (Um sicherzustellen, dass dies geschieht, hätte man auch nur die Subroutinen-
    ! Variante exportieren können.) Dann kann deleteAll normal benutzt werden.
    
 contains 

    integer function square(i)! quadrieren
        integer, intent(in) :: i
        square = i**2
    end function square

    integer function add(a, b)! addieren
        integer, intent(in) :: a, b 
        add = a+b
    end function

    logical function even(i)! testen, ob gerade
        integer, intent(in) :: i
        even= mod(i,2)==0
    end function even
    
end program main